@@ -1,8 +1,9 @@
 package lib::filter;
 
-our $DATE = '2015-07-01'; # DATE
-our $VERSION = '0.20'; # VERSION
+our $DATE = '2016-07-09'; # DATE
+our $VERSION = '0.25'; # VERSION
 
+# no need to avoid strict & warnings, because Config uses them
 use strict;
 use warnings;
 use Config;
@@ -39,7 +40,7 @@ sub _open_handle {
 }
 
 my $hook;
-my $orig_inc;
+my ($orig_inc, $orig_inc_sorted_by_len);
 
 sub import {
     my ($class, %opts) = @_;
@@ -53,6 +54,7 @@ sub import {
                         allow_core|allow_noncore|
                         extra_inc|
                         allow|allow_list|allow_re|
+                        allow_is_recursive|
                         disallow|disallow_list|disallow_re|
                         filter
                     )\z/x;
@@ -71,7 +73,10 @@ sub import {
         unshift @INC, split(/:/, $opts{extra_inc});
     }
 
-    $orig_inc ||= [@INC];
+    if (!$orig_inc) {
+        $orig_inc = [@INC];
+        $orig_inc_sorted_by_len = [sort {length($b) <=> length($a)} @INC];
+    }
 
     my $core_inc = [@Config{qw(privlibexp archlibexp)}];
     my $noncore_inc = [grep {$_ ne $Config{privlibexp} &&
@@ -96,6 +101,7 @@ sub import {
     if ($opts{disallow}) {
         for (split /\s*;\s*/, $opts{disallow}) {
             $disallow{$_} = "disallow";
+            (my $pm = "$_.pm") =~ s!::!/!g; delete $INC{$pm};
         }
     }
     if ($opts{disallow_list}) {
@@ -105,42 +111,69 @@ sub import {
             $line =~ s/^\s+//;
             $line =~ /^(\w+(?:::\w+)*)/ or next;
             $disallow{$1} ||= "disallow_list";
+            (my $pm = "$1.pm") =~ s!::!/!g; delete $INC{$pm};
         }
     }
 
     $hook = sub {
         my ($self, $file) = @_;
 
-        warn "$dbgh hook called for $file\n" if $opts{debug};
+        my @caller = caller(0);
+
+        warn "$dbgh hook called for $file (from package $caller[0] file $caller[1])\n" if $opts{debug};
 
         my $path;
       FILTER:
         {
             my $mod = $file; $mod =~ s/\.pm$//; $mod =~ s!/!::!g;
+            my $err_prefix = "Can't locate $file";
             if ($opts{filter}) {
                 local $_ = $mod;
                 warn "$dbgh Checking against custom filter ...\n" if $opts{debug};
                 unless ($opts{filter}->($mod)) {
-                    die "Module '$mod' is disallowed (filter)";
+                    die "$err_prefix (module '$mod' is disallowed (filter))";
                 }
             }
             if ($opts{disallow_re} && $mod =~ /$opts{disallow_re}/) {
-                die "Module '$mod' is disallowed (disallow_re)";
+                die "$err_prefix (module '$mod' is disallowed (disallow_re))";
             }
             if ($disallow{$mod}) {
-                die "Module '$mod' is disallowed ($disallow{$mod})";
+                die "$err_prefix (module '$mod' is disallowed ($disallow{$mod}))";
             }
             if ($opts{allow_re} && $mod =~ /$opts{allow_re}/) {
                 warn "$dbgh module $mod matches allow_re\n" if $opts{debug};
                 $path = module_path($file, $orig_inc);
                 last FILTER if $path;
-                die "Module '$mod' is allowed (allow_re) but can't locate $file in \@INC (\@INC contains: ".join(" ", @INC);
+                die "$err_prefix (module '$mod' is allowed (allow_re) but can't locate $file in \@INC (\@INC contains: ".join(" ", @INC)."))";
             }
             if ($allow{$mod}) {
                 warn "$dbgh module $mod matches $allow{$mod}\n" if $opts{debug};
                 $path = module_path($file, $orig_inc);
                 last FILTER if $path;
-                die "Module '$mod' is allowed ($allow{$mod}) but can't locate $file in \@INC (\@INC contains: ".join(" ", @INC);
+                die "$err_prefix (module '$mod' is allowed ($allow{$mod}) but can't locate $file in \@INC (\@INC contains: ".join(" ", @INC)."))";
+            }
+            if ($opts{allow_is_recursive}) {
+                my $caller_pkg_from_file;
+                for (@$orig_inc_sorted_by_len) {
+                    #print "D:\$_=<$_> vs $caller[1]\n";
+                    if (index($caller[1], $_) == 0) {
+                        $caller_pkg_from_file = substr($caller[1], length($_)+1);
+                        #print "D:caller_pkg_from_file=<$caller_pkg_from_file>\n";
+                        $caller_pkg_from_file =~ s/\.pm\z//;
+                        $caller_pkg_from_file =~ s/\Q$Config{archname}\E.//;
+                        $caller_pkg_from_file =~ s![/\\]!::!g;
+                    }
+                }
+                for my $caller_pkg (grep {defined} $caller[0], $caller_pkg_from_file) {
+                    (my $pm = "$caller_pkg.pm") =~ s!::!/!g;
+                    if (exists $INC{$pm}) {
+                        $path = module_path($file, $orig_inc);
+                        if ($path) {
+                            warn "$dbgh module '$mod' allowed because it is require'd by $caller_pkg (allow_is_recursive=1)\n" if $opts{debug};
+                            last FILTER;
+                        }
+                    }
+                }
             }
 
             my $inc;
@@ -207,7 +240,7 @@ lib::filter - Allow/disallow loading modules
 
 =head1 VERSION
 
-This document describes version 0.20 of lib::filter (from Perl distribution lib-filter), released on 2015-07-01.
+This document describes version 0.25 of lib::filter (from Perl distribution lib-filter), released on 2016-07-09.
 
 =head1 SYNOPSIS
 
@@ -241,7 +274,7 @@ distribution.
 
 =item * To only allow a specific set of modules:
 
- % perl -Mlib::filter=allow_core,0,allow_noncore,0,allow,'XSLoader,List::Util' yourscript.pl
+ % perl -Mlib::filter=allow_core,0,allow_noncore,0,allow,'XSLoader;List::Util' yourscript.pl
 
 =item * To allow core modules plus some additional modules:
 
@@ -254,7 +287,7 @@ For example to test a fatpacked script that might still require some XS modules:
  % perl -Mlib::filter=allow_noncore,0,allow_list,'/tmp/allow.txt' yourscript.pl
 
  # allow core modules plus additional modules found in some dirs
- % perl -Mlib::filter=allow_noncore,0,extra_path,'.:proj/lib' yourscript.pl
+ % perl -Mlib::filter=allow_noncore,0,extra_inc,'.:proj/lib' yourscript.pl
 
  # allow some non-core XS modules
  % perl -MModule::CoreList -Mlib::filter=filter,'sub{ return 1 if Module::CoreList->is_core($_); return 1 if m!Clone|Acme/Damn!; 0' yourscript.pl
@@ -263,7 +296,14 @@ For example to test a fatpacked script that might still require some XS modules:
 Alternatively, you can also test by preloading the additional modules before
 using lib::core::only:
 
- % perl -MClone -MAcme::Damn -Mlib::core::only yourscript.pl
+ % perl -mClone -mAcme::Damn -Mlib::core::only yourscript.pl
+
+=item * To allow a module and recursively all other modules that the module requires
+
+This is convenient when we want to allow a non-trivial module which itself uses
+some other modules, e.g. L<Moo> or L<Moose>:
+
+ % perl -Mlib::filter=allow_noncore,0,allow,Moo,allow_is_recursive=0
 
 =item * To disallow some modules:
 
@@ -321,6 +361,14 @@ Allow modules matching regex pattern.
 
 Read a file containing list of modules to allow (one module per line).
 
+=item * allow_is_recursive => bool (default: 0)
+
+If set to 1, then will also allow modules that are required by the allowed
+modules (and modules that are allowed by I<those> modules, and so on). This is
+convenient if you want to allow a non-trivial module, say, L<Moo> or L<Moose>
+which will require other modules too. Without this option, you will need to
+explicitly allow each of those modules yourself.
+
 =item * allow_core => bool (default: 1)
 
 Allow core modules.
@@ -347,15 +395,22 @@ How a module is filtered:
 
 =item * First it's checked against C<filter>, if that option is defined
 
-=item * then, it is checked against the disallow/disallow_re/disallow_list.
+=item * then, it is checked against the C<disallow>/C<disallow_re>/C<disallow_list>.
 
 If it matches one of those options then the module is disallowed.
 
-=item * Otherwise it is checked against the allow/allow_re/allow_list.
+=item * Otherwise it is checked against the C<allow>/C<allow_re>/C<allow_list>.
 
 If it matches one of those options and the module's path is found in the
 directories in C<@INC>, then the module is allowed.
 
+=item * If C<allow_is_recursive> is true, check the requirer.
+
+If the calling package is already in C<%INC>, we allow that. For example, if we
+allow C<Moo> and C<Moo> calls L<Moo::_strictures> and L<Module::Runtime>, we
+will also allow them. Later if C<Moo::_strictures> tries to load L<strictures>,
+we also allow it, and so on.
+
 =item * Finally, allow_core/allow_noncore is checked.
 
 When C<allow_core> is set to false, core directories are excluded. Likewise,
@@ -363,14 +418,6 @@ when C<allow_noncore> is set to false, non-core directories are excluded.
 
 =back
 
-=head1 SEE ALSO
-
-L<lib::none>
-
-L<lib::core::only>
-
-L<Devel::Hide>
-
 =head1 HOMEPAGE
 
 Please visit the project's homepage at L<https://metacpan.org/release/lib-filter>.
@@ -387,13 +434,21 @@ When submitting a bug or request, please include a test-file or a
 patch to an existing test-file that illustrates the bug or desired
 feature.
 
+=head1 SEE ALSO
+
+Related/similar modules: L<lib::none>, L<lib::core::only>, L<Devel::Hide>,
+L<Test::Without::Module>.
+
+To simulate the absence of certain programs in PATH, you can try
+L<File::Which::Patch::Hide>.
+
 =head1 AUTHOR
 
 perlancar <perlancar@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2015 by perlancar@cpan.org.
+This software is copyright (c) 2016 by perlancar@cpan.org.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
